/*
Copyright (C) 2014 Graham Jones, www.indriid.com
*/

package stacey.util;

import beast.core.Description;
import beast.evolution.tree.Node;
import beast.evolution.tree.TreeInterface;
import beast.util.Randomizer;

import java.util.*;

/**
 * Created by Graham Jones on 26/10/2014.
 */



/*
This is for operators to do calculations. It is sort of
parallel to PIOProbabilitySMCTree, but it has much less data in it.
But it does have references to BEAST tree nodes so that oce the
operator has calculated what to do, it can make it happen.

A OperatorsSMCTree is created by the operator, used, and discarded.

*/

@Description("A tree built from the underlying SMC-tree which is used by operators to choose details of moves")
public class OperatorsSMCTree {

    private final TreeInterface tree;
    private final Bindings bindings;
    private SMCUnionNode[] opSNodes;

    /*
    *  The anc, lft, rgt, height, ID fields implement the basic binary tree.
    *  The other fields are 'working' fields.
    *
    *  union represents a union of sets (minimal clusters) of individuals.
    *  These are used for deciding details of moves.
    *  beasttreenode refers to underlying BEAST Tree
    */
    public class SMCUnionNode implements UnionNode {
        private int anc;
        private int lft;
        private int rgt;
        private double height;
        private String ID;
        private final BitUnion union;
        private Node beasttreenode;


        // constructor that claims memory, makes 'empty' node
        SMCUnionNode(int nTips) {
            anc = -1;
            lft = -1;
            rgt = -1;
            height = -1.0;
            ID = "";
            union = new BitUnion(nTips);
            beasttreenode = null;
        }


        public int getLft() { return lft; }
        public int getRgt() { return rgt; }
        public double getHeight() { return height;}
        public BitUnion getUnion() { return union; }
        public String getID() { return ID; }
        public String getExtraText() { return  (beasttreenode==null ? "null" : "has beast node");  }

        public void setLft(int lft) { this.lft = lft; }
        public void setRgt(int rgt) { this.rgt = rgt; }
        public void setAnc(int anc) { this.anc = anc; }
        public void setHeight(double height) { this.height = height; }
        public void setBNode(Node bNode) { beasttreenode = bNode; }

        public void setTipUnionAndID(String id) {
            this.ID = id;
            union.replaceWith(bindings.tipUnionFromSmcID(id));
        }


        void reset() {
            anc = -1;
            lft = -1;
            rgt = -1;
            height = -1.0;
            ID = "";
            union.reset();
            beasttreenode = null;
        }

    } // end PopsIONode


    public static class SPRmoveInfo {
        final BitUnion sppS;
        final BitUnion sppDs[];
        final double heightsDancs[];

        SPRmoveInfo(BitUnion sppS,
                    BitUnion [] sppDs,
                    double [] heightsDancs) {
            this.sppS = sppS;
            this.sppDs = sppDs;
            this.heightsDancs = heightsDancs;
    }
    }


    /***********************************************************************************/
    /************************** main class *********************************************/
    /***********************************************************************************/

    // constructor which claims memory
    public OperatorsSMCTree(TreeInterface tree, Bindings bindings) {
        this.tree = tree;
        this.bindings = bindings;
        int nTips = bindings.smcTreeTipCount();
        assert nTips == tree.getTaxonset().getTaxonCount();
        assert nTips == tree.getLeafNodeCount();
        int nNodes = 2*nTips-1;
        opSNodes = new SMCUnionNode[2*nTips-1];
        for (int n = 0; n < nNodes; n++) {
            opSNodes[n] = new SMCUnionNode(nTips);
        }
        reset();
        }


    // update from BEAST tree
    public void update() {
        assert opSNodes[0].beasttreenode == null;
        UnionNode.Abstract.beastTreeToUnionTree(opSNodes, tree);
    }


    public void reset() {
        for (int n = 0; n < opSNodes.length; n++) {
            opSNodes[n].reset();
        }
    }


    public int siblingOf(int x) {
        int a = opSNodes[x].anc;
        assert a >= 0;
        int s;
        if (opSNodes[a].lft == x) { s = opSNodes[a].rgt; } else { s = opSNodes[a].lft; }
        return s;
    }


    public BitUnion getUnion(int n) {
        return opSNodes[n].getUnion();
    }


    /******************************  debug  ********************************************/

    public SMCUnionNode [] getNodes() {
        return Arrays.copyOf(opSNodes, opSNodes.length);
    }



    public String toString() {
        return UnionNode.Abstract.treeAsText(opSNodes, "beast_node");
    }




    /********************************* NodesNudge *********************************************/

    public NudgeMoveSMCNodeInfo chooseNodeForNudge() {
        int s;
        do {
            s = Randomizer.nextInt(opSNodes.length);
        } while (opSNodes[s].lft < 0  ||  opSNodes[s].anc < 0);
        int sl = opSNodes[s].lft;
        int sr = opSNodes[s].rgt;
        int sa = opSNodes[s].anc;
        double shgt = opSNodes[s].height;
        BitUnion sunion = opSNodes[s].union;
        BitUnion lftunion = opSNodes[sl].union;
        BitUnion rgtunion = opSNodes[sr].union;
        double lfthgt = opSNodes[sl].height;
        double rgthgt = opSNodes[sr].height;
        double anchgt = (sa >= 0) ? opSNodes[sa].height : Double.POSITIVE_INFINITY;
        return new NudgeMoveSMCNodeInfo(s, shgt, sunion, lfthgt, lftunion, rgthgt, rgtunion, anchgt);
    }


    public double doNodeNudge(int s, double dh) {
        opSNodes[s].beasttreenode.setHeight(opSNodes[s].beasttreenode.getHeight() + dh);
        return 0.0;
    }


    /****************************** Focused scaler ******************************************/

    public FSMoveSMCNodeUnions setUpForFSMove(double [] logHeights, int [] distances) {
        fillInLogHeights(logHeights);
        int s;
        do {
            s = Randomizer.nextInt(opSNodes.length);
        } while (!nodeCanBeFocus(s));
        int sl = opSNodes[s].lft;
        int sr = opSNodes[s].rgt;
        BitUnion sunion = opSNodes[s].union;
        BitUnion lftunion = opSNodes[sl].union;
        BitUnion rgtunion = opSNodes[sr].union;
        fillInDistances(s, distances);
        return new FSMoveSMCNodeUnions(sunion, lftunion, rgtunion);
    }


    private boolean nodeCanBeFocus(int s) {
        // not a tip, not the root, and at least one child not a tip
        if (opSNodes[s].lft < 0) {
            return false;
        }if (opSNodes[s].anc < 0) {
            return false;
        }
        int lft = opSNodes[s].lft;
        int rgt = opSNodes[s].rgt;
        return (opSNodes[lft].lft >= 0 || opSNodes[rgt].lft >= 0);
    }

    // h() is logHeight, w() is weight. n is node, a is parent. x is amount to move.
    // This must be true after move:
    // h(a) + w(a)*x >= h(n) + w(n)*x
    // h(a) - h(n) >= (w(n) - w(a))x
    // If w(n) - w(a) > 0, get x <= (h(a) - h(n))/(w(n) - w(a))
    // If w(n) - w(a) < 0, get x >= (h(a) - h(n))/(w(n) - w(a))
    public void getFocusedScalerBounds(double[] interval, double[] logHeights, double[] weights) {
        for (int n = 0; n < opSNodes.length; n++) {
            if (opSNodes[n].lft >= 0  &&  opSNodes[n].anc >= 0) {
                int anc = opSNodes[n].anc;
                double weightDiff = weights[n] - weights[anc];
                double loghDiff = logHeights[anc] - logHeights[n];
                if (weightDiff > 0) {
                    interval[1] = Math.min(interval[1], loghDiff/weightDiff);
                }
                if (weightDiff < 0) {
                    interval[0] = Math.max(interval[0], loghDiff/weightDiff);
                }
            }
        }
    }



    public  FSMoveSMCNodeInfo[] getAllInternalFSMoveSNInfos(double[] logHeights, double[] weights) {
        FSMoveSMCNodeInfo[] fsNInfos = new FSMoveSMCNodeInfo[(opSNodes.length - 1)/ 2];
        int i = 0;
        for (int s = 0; s < opSNodes.length; s++) {
            if (opSNodes[s].lft >= 0) {
                assert i < fsNInfos.length;
                int sl = opSNodes[s].lft;
                int sr = opSNodes[s].rgt;
                BitUnion sunion = opSNodes[s].union;
                BitUnion lftunion = opSNodes[sl].union;
                BitUnion rgtunion = opSNodes[sr].union;
                double logHeight = logHeights[s];
                double weight = weights[s];
                fsNInfos[i++] = new FSMoveSMCNodeInfo(sunion, lftunion, rgtunion, logHeight, weight);
            }
        }
        assert i == fsNInfos.length;
        return fsNInfos;
    }


    // this actually edits the BEAST tree
    public double fsMoveDoScale(double logSF, double [] weights) {
        double logHR = 0.0;
        for (int s = 0; s < opSNodes.length; s++) {
            if (opSNodes[s].lft >= 0 && opSNodes[s].anc >= 0 && weights[s] > 0.0) {
                double logSFW = logSF * weights[s];
                opSNodes[s].beasttreenode.setHeight(opSNodes[s].beasttreenode.getHeight() * Math.exp(logSFW));
                logHR += logSFW;
            }
        }
        return logHR;
    }


    /***************    CoordinatedPruneRegraft  ***************************************/
    /*
    1. Choose a node s (source). chooseSubtreeForSPR()
    2. Choose a node d (destination) so that d's branch contains time of s. chooseDestinationForSPR()
    3. Find info the gene trees need to know. makeSPRmoveInfo()
    4. Move the subtree with root s into branch d. doSPRmove()

         s                        .                         s
    x   /          d              .              x           \   d
     \ /          /               .               \           \ /
      p          /                .                \           p
       \        /                 .                 \         /
      --y      /                  .                --y       /
         \    z--e0               .                   \    z--
          \ /                     .                    \ /
           m = d2                 .                     m

    The gene trees need to know
    sppS the tips of s
    hM   the height of m, the mrca of s and d,
    sppD the tips of d
    sppEs the tips of the side branches between d and m, just e0 here
    heightsEancs the heights of the parents of these, just height of z here
    */

    // choose a random source node s, not root, not child of root.
    // If NNI, restrict to s with appropriate cousin.
    public int chooseSubtreeForSPR(boolean NNI) {
        assert opSNodes.length > 3;
        int s;
        if (NNI) {
            do {
                s = Randomizer.nextInt(opSNodes.length);
            }
            while (rejectSubtreeForNNI(s));
        } else {
            do {
                s = Randomizer.nextInt(opSNodes.length);
            }
            while (rejectSubtree(s));
        }
        return s;
    }



    private boolean rejectSubtree(int s) {
        return (s == opSNodes.length-1  ||  opSNodes[s].anc == opSNodes.length-1);
    }


    private boolean rejectSubtreeForNNI(int s) {
        if (s == opSNodes.length-1  ||  opSNodes[s].anc == opSNodes.length-1) {
            return true;
        }
        int anc = opSNodes[s].anc;
        int d = siblingOf(anc);
        return  (opSNodes[d].height >= opSNodes[anc].height);
    }





    // choose a destination for subtree s.  debugnb[1] is for debugging/tuning
    public int chooseDestinationForSPR(int s, boolean NNI, Integer [] debugnb) {
        if (NNI) {
            int anc =  opSNodes[s].anc;
            debugnb[0] = 3;
            return siblingOf(anc);
        } else {
            int sib = siblingOf(s);
            int anc =  opSNodes[s].anc;
            double h = opSNodes[anc].height;
            ArrayList<Integer> dests = new ArrayList<Integer>(1 + opSNodes.length/2);
            ArrayList<Integer> nbs = new ArrayList<Integer>(1 + opSNodes.length/2);
            for (int n = 0; n < opSNodes.length; n++) {
                if (n != s  &&  n != anc  &&  n != sib &&
                        opSNodes[n].height <= h  &&
                        opSNodes[ opSNodes[n].anc ].height >= h) {
                    dests.add(n);
                    int nb = debugNumberOfBranchesBetween(s, n);
                    assert nb >= 3;
                    nbs.add(nb);

                }
            }
            int choice = Randomizer.nextInt(dests.size());
            int d = dests.get(choice);
            debugnb[0] = nbs.get(choice);
            return d;
        }
    }



    // Gather information that gene trees need to make moves corresponding to s->d
    public SPRmoveInfo makeSPRmoveInfo(int s, int d) {
        BitUnion sppS = opSNodes[s].union;
        int m = mrcaOfPair(s, d);
        ArrayList<Integer> ds = new ArrayList<Integer>();
        while (d != m) {
            ds.add(d);
            d = opSNodes[d].anc;
        }
        BitUnion [] sppDs = new BitUnion[ds.size()];
        double [] heightsDancs = new double[ds.size()];
        for (int i = 0; i < ds.size(); i++) {
            int di = ds.get(i);
            sppDs[i] = opSNodes[di].union;
            heightsDancs[i] = opSNodes[ opSNodes[di].anc ].height;
        }
        SPRmoveInfo smi = new SPRmoveInfo(sppS, sppDs, heightsDancs);
        return smi;
    }


    // this one actually edits the smc tree
    public void doSPRmove(int s, int d) {
        Node sr = opSNodes[s].beasttreenode;
        Node dr = opSNodes[d].beasttreenode;
        Node ar = sr.getParent();
        Node xr = ar.getChild(0) == sr ? ar.getChild(1) : ar.getChild(0);
        Node yr = ar.getParent();
        Node zr = dr.getParent();
        zr.removeChild(dr);
        ar.removeChild(xr);
        yr.removeChild(ar);
        zr.addChild(ar);
        ar.addChild(dr);
        yr.addChild(xr);
    }


    // for findind indices of s and x after move so can calculate reverse HR
    public int indexFromUnion(BitUnion union) {
        return UnionNode.Abstract.nodeIndexOfUnionInSubtree(opSNodes, opSNodes.length - 1, union);
    }


    /*************************************************************************************/
    /*****************************     private     ***************************************/
    /*************************************************************************************/

    /*********************** private for debugging  *********************************/

    private int debugNumberOfBranchesBetween(int x, int y) {
        int n = 0;
        while (x != y) {
            if (opSNodes[x].height < opSNodes[y].height) {
                x = opSNodes[x].anc;
            } else {
                y = opSNodes[y].anc;
            }
            n++;
        }
        return n;
    }


    /****************** private FocusedNodeHeightScaler ****************************************/

    private void fillInDistances(int n, int [] distances) {
        assert opSNodes.length == distances.length;
        for (int i = 0; i < distances.length; i++) {
            distances[i] = Integer.MAX_VALUE;
        }
        distances[n] = 0;
        while (opSNodes[n].anc >= 0) {
            int anc = opSNodes[n].anc;
            distances[anc] = Math.min(distances[anc], distances[n] + 1);
            n = anc;
        }
        fillInSubtreeDistances(opSNodes.length-1, distances);
    }


    private void fillInSubtreeDistances(int n, int [] distances) {
        if (opSNodes[n].anc >= 0) {
            int anc = opSNodes[n].anc;
            distances[n] = Math.min(distances[n], distances[anc] + 1);
        }
        if (opSNodes[n].lft >= 0) {
            fillInSubtreeDistances(opSNodes[n].lft, distances);
            fillInSubtreeDistances(opSNodes[n].rgt, distances);
        }
    }


    private void fillInLogHeights(double [] logHeights) {
        assert opSNodes.length == logHeights.length;
        for (int i = 0; i < opSNodes.length; i++) {
            if (opSNodes[i].lft >= 0) {
                logHeights[i] = Math.log(opSNodes[i].height);
            } else {
                logHeights[i] = Double.NEGATIVE_INFINITY;
            }
        }
    }



    /****************** private: CoordinatedPruneRegraft ************************/

    private int mrcaOfPair(int x, int y) {
        while ( x != y) {
            if (opSNodes[x].height < opSNodes[y].height) {
                x = opSNodes[x].anc;
            } else {
                y = opSNodes[y].anc;
            }
        }
        return x;
    }

}
