/*
Copyright (C) 2014 Graham Jones, www.indriid.com
*/
package stacey.util;

import beast.core.Description;
import beast.evolution.tree.Node;
import beast.evolution.tree.Tree;

import java.util.*;

/**
 * Created by Graham Jones on 26/10/2014.
 */

/*
This is for P(smcCoalescent) calculations. It is sort of
parallel to OperatorsGeneTree. It cooperates with PIOProbabilitySMCTree.

It has an inner classes, GeneUnionNode. T

The main methods are
fitsInSpeciesTree()
fillSpeciesTreeWithGeneCoalescentInfo()
both called from PIOMSCoalescentDistribution.
They both make a GeneUnionTree, then visit the nodes and from each one,
call PIOProbabilitySMCTree methods, discard the GeneUnionTree.




 */

@Description("A tree built from an underlying gene tree which is used in calculation of P(smcCoalescent)")

public class PIOProbabilityGeneTree {

    public final Tree tree;
    private final double coalFactor;
    private final int index;
    private final Bindings bindings;

    private final GeneUnionNode[] pioGNodes;


    // Node for PIOGeneTree.GeneUnionTree.
    public class GeneUnionNode implements UnionNode {
        private int lft;
        private int rgt;
        private double height;
        private final BitUnion union;
        private String ID; // for debugging


        // Constructor makes a half-formed tip node.
        public GeneUnionNode(int smcTreeLeafCount) {
            lft = -1;
            rgt = -1;
            height = 0.0;
            union = new BitUnion(smcTreeLeafCount);
            ID = "";
        }


        public int getLft() { return lft; }
        public int getRgt() { return rgt; }
        public double getHeight() { return  height; }
        public BitUnion getUnion() { return  union; }
        public String getID() { return ID; }
        public String getExtraText() { return ""; }


        public void setLft(int lft) { this.lft = lft; }
        public void setRgt(int rgt) { this.rgt = rgt; }
        public void setAnc(int anc) { /* nothing to do */ }
        public void setHeight(double height) { this.height = height; }
        public void setBNode(Node bNode) { /* nothing to do */ }

        public void setTipUnionAndID(String id) {
            this.ID = id;
            assert id != null;
            assert bindings != null;
            union.replaceWith(bindings.tipUnionFromAllTaxaID(id));
            /*int ati = bindings.allTaxaID2index(id);
            int smc = bindings.allTaxaIndex2smcIndex(ati);
            union.insert(smc);*/
        }

        public void reset() {
            lft = -1;
            rgt = -1;
            height = 0.0;
            union.reset();
            ID = "";
        }

    }



    /************************** main class *****************************************/



    // Constructor which reserves memory
    public PIOProbabilityGeneTree(Tree tree, double coalFactor, Bindings bindings, int j) {
        this.tree = tree;
        this.coalFactor = coalFactor;
        this.bindings = bindings;
        this.index = j;
        pioGNodes = new GeneUnionNode[tree.getNodeCount()];
        for (int i = 0; i < pioGNodes.length; i++) {
            pioGNodes[i] = new GeneUnionNode(bindings.smcTreeTipCount());
        }
    }


    // updates the gene-union tree from its BEAST tree
    public void update() {
        for (int n = 0; n < pioGNodes.length; n++) {
            pioGNodes[n].reset();
        }
        UnionNode.Abstract.beastTreeToUnionTree(pioGNodes, tree);
    }


    public boolean fitsInSpeciesTree(PIOProbabilitySMCTree smcTree) {
        update();
        boolean fits = subtreeFitsInSpeciesTree(pioGNodes.length - 1, smcTree);
        return fits;
    }


    //Sets things up for likelihood calculation with info from one gene
    public void fillSpeciesTreeWithGeneCoalescentInfo(PIOProbabilitySMCTree smcTree, double maxgtreehgt) {
        update();
        recordCoalescences(smcTree);
        smcTree.fillInCoalCountsIntensities(maxgtreehgt, coalFactor, index);
    }



    /*******************************  debug  **********************************************/


    public GeneUnionNode [] getNodes() {
        return Arrays.copyOf(pioGNodes, pioGNodes.length);
    }



    public String toString() {
        return UnionNode.Abstract.treeAsText(pioGNodes, "");
    }


    /************************ private ***************************************/



    private void recordCoalescences(PIOProbabilitySMCTree smcTree) {
        smcTree.clearCoalescences(index);
        subtreeRecordCoalescences(pioGNodes.length - 1, smcTree);
    }



    private boolean subtreeFitsInSpeciesTree(int n, PIOProbabilitySMCTree smcTree) {
        if (pioGNodes[n].lft >= 0) {
            if (!subtreeFitsInSpeciesTree(pioGNodes[n].lft, smcTree)) {
                return false;
            }
            if (!subtreeFitsInSpeciesTree(pioGNodes[n].rgt, smcTree)) {
                return false;
            }
        }
        return smcTree.coalescenceIsCompatible(pioGNodes[n].height, pioGNodes[n].union);
    }


    private void subtreeRecordCoalescences(int n, PIOProbabilitySMCTree smcTree) {
        if (pioGNodes[n].lft >= 0) {
            subtreeRecordCoalescences(pioGNodes[n].lft, smcTree);
            subtreeRecordCoalescences(pioGNodes[n].rgt, smcTree);
            smcTree.recordCoalescence(pioGNodes[n].height, pioGNodes[n].union, index);
        }
    }


    private void subtreedebugCheckCoalescences(int n, PIOProbabilitySMCTree smcTree) {
        if (pioGNodes[n].lft >= 0) {
            subtreedebugCheckCoalescences(pioGNodes[n].lft, smcTree);
            subtreedebugCheckCoalescences(pioGNodes[n].rgt, smcTree);
            smcTree.debugCheckCoalescenceIsThere(pioGNodes[n].height, pioGNodes[n].union, index);
        }
    }

}