/*
Copyright (C) 2014 Graham Jones, www.indriid.com
*/
package stacey.util;

import beast.core.Description;
        import beast.evolution.tree.Node;
        import beast.evolution.tree.TreeInterface;

import java.util.*;

/**
 * Created by Graham Jones on 26/10/2014.
 */


/*
This is an intermediate result in the sense of BEAST2 book 14.4.3. It supports the
calculation of the smcCoalescent by PIOMSCoalescentDistribution. It has a
constructor which is called twice to reserve memory. Once for the current
smcTree, and once for the stored or old version. It has a update() method
for (re)creating itself from the underlying BEAST tree when that changes.
It has a replaceWith() method for storing itself before a move.

The data is in arrays pioSNodes, gtreeCoalescences.
The former stores a copy of the SMC-Tree topology and nodes heights with
extra information. The is updated from gene trees, and then the information
in nodes is updated from this.

There are four inner classes: Coalescence, GTreeCoalescences,
PerBranchPerGeneInfo, PopsIONode.

A Coalescence is a single coalescence from one gene tree.
GTreeCoalescences is an array of them for one gene tree.
Field gtreeCoalescences[] is an array of these.

PerBranchPerGeneInfo is a part of a PopsIONode. Field pioSNodes[]
is an array of PopsIONode which implement the  tree.

*/

@Description("A tree built from the underlying SMC-tree which is used store intermediate results in calculation of P(smcCoalescent)")

public class PIOProbabilitySMCTree {

    private final TreeInterface tree;
    private final Bindings bindings;
    private final PopsIONode[] pioSNodes;
    private final GTreeCoalescences gtreeCoalescences[];



    /****************** two inner classes for GTreeCoalescences *******************************/

    private class Coalescence {
        double height;
        final BitUnion union;
        int nn;  // index in pioSNodes[]

        // constructor that reserves memory
        Coalescence() {
            height = -1.0;
            union = new BitUnion(bindings.smcTreeTipCount());
            nn = -1;
        }

        void setFrom(double height, BitUnion union) {
            this.height = height;
            this.union.replaceWith(union);
            nn = UnionNode.Abstract.nodeIndexOfUnionInSubtree(pioSNodes, pioSNodes.length - 1, union);

            assert (pioSNodes[nn].height <= height);
            int anc = pioSNodes[nn].anc;
            while (anc >= 0  &&  pioSNodes[anc].height <= height) {
                nn = anc;
                anc = pioSNodes[nn].anc;
            }
        }

        void replaceWith(Coalescence x) {
            height = x.height;
            union.replaceWith(x.union);
            nn = x.nn;
        }
    }





    private class GTreeCoalescences {
        final int j; // gtree index
        final ArrayList<Coalescence> coalescences;
        int nextn;

        // constructor that reserves memory
        GTreeCoalescences(int j, int ncoals) {
            this.j = j;
            coalescences = new ArrayList<Coalescence>(ncoals);
            for (int c = 0; c < ncoals; c++) {
                coalescences.add(new Coalescence());
            }
            nextn = 0;
        }


        void clearCoalescences() {
            nextn = 0;
        }

        void addCoalescence(double height, BitUnion union) {
            assert nextn < coalescences.size();
            coalescences.get(nextn).setFrom(height, union);
            nextn++;
        }


        void sortCoalescences() {
            Collections.sort(coalescences, COALESCENCE_ORDER);
        }


        void setPIONodeCoalHeights() {
            for (int n = 0; n < pioSNodes.length; n++) {
                pioSNodes[n].geneInfo[j].ch0 = -1;
                pioSNodes[n].geneInfo[j].ch1 = -1;
            }
            for (int c = 0; c < nextn; c++) {
                int nn = coalescences.get(c).nn;
                if (c == 0  ||  nn != coalescences.get(c-1).nn) {
                    pioSNodes[nn].geneInfo[j].ch0 = c;
                }
                if (c == nextn-1  ||  nn != coalescences.get(c+1).nn) {
                    pioSNodes[nn].geneInfo[j].ch1 = c+1;
                }
            }
        }


        void replaceWith(GTreeCoalescences x) {
            assert  j == x.j;
            nextn = x.nextn;
            for (int c = 0; c < coalescences.size(); c++) {
                coalescences.get(c).replaceWith(x.coalescences.get(c));
            }
        }

    }


    /****************** two inner classes for PIONodes *******************************/


    private static class PerBranchPerGeneInfo {
        private int ch0;       // ch0 and ch1 define a range in coalescences[].height[]
        private int ch1;       // providing t_1, ... t_K, where K = k_{jb} = ch1-ch0
        private int nlineages;                        // n_{jb}
        private double minusLogCoalFactor;            // - k_{jb} log(p_j)
        private double coalintensity;                 // j'th part of gamma_b, namely
                                                      // (1/p_j) sum_0^K c_{jbi} ((n_{jb} - 1) choose 2)

        // dud constructor
        PerBranchPerGeneInfo() {
            ch0 = -1;
            ch1 = -1;
            nlineages = 0;
            minusLogCoalFactor = 0.0;
            coalintensity = 0.0;
        }

        void reset() {
            ch0 = -1;
            ch1 = -1;
            nlineages = 0;
            minusLogCoalFactor = 0.0;
            coalintensity = 0.0;
        }


        void replaceWith(PerBranchPerGeneInfo x) {
            ch0 = x.ch0;
            ch1 = x.ch1;
            nlineages = x.nlineages;
            minusLogCoalFactor = x.minusLogCoalFactor;
            coalintensity = x.coalintensity;
        }

    }


    /*
    *  The anc, lft, rgt, height, ID fields implement the basic binary tree.
    *  The other fields are 'working' fields.
    *
    *  union represents a union of sets (minimal clusters) of individuals.
    *  These are used for calculations: determining compatibility
    *  with gene trees; calculating P(smcCoalescent).
    *  geneInfo[] stores coalescent information from all gene trees. It
    *  contains intermediate values for calculating P(smcCoalescent).
    */
    private class PopsIONode implements UnionNode {
        private int anc;
        private int lft;
        private int rgt;
        private double height;
        private String ID;
        private final BitUnion union;
        private final PerBranchPerGeneInfo [] geneInfo;


        // constructor that reserves memory
        PopsIONode(int nGTrees, int nTips) {
            anc = -1;
            lft = -1;
            rgt = -1;
            height = -1.0;
            ID = "";
            union = new BitUnion(nTips);
            geneInfo = new PerBranchPerGeneInfo[nGTrees];
            for (int j = 0; j < nGTrees; j++) {
                geneInfo[j] = new PerBranchPerGeneInfo();
            }
        }


        public int getLft() { return lft; }
        public int getRgt() { return rgt; }
        public double getHeight() { return  height; }
        public BitUnion getUnion() { return  union; }
        public String getID() { return ID; }

        public String getExtraText() {
            StringBuilder s = new StringBuilder();
            Formatter formatter = new Formatter(s, Locale.US);

            for (int j = 0; j < geneInfo.length; j++) {
                formatter.format("%3d  ", geneInfo[j].nlineages);
                if (geneInfo[j].ch0 == geneInfo[j].ch1) {
                    formatter.format("no coals; ");
                }
                if (geneInfo[j].ch0 + 1 == geneInfo[j].ch1) {
                    formatter.format("one at ");
                    formatter.format(Misc.nonnegIn8Chars(gtreeCoalescences[j].coalescences.get(geneInfo[j].ch0).height) + "; ");
                }
                if (geneInfo[j].ch0 + 1 < geneInfo[j].ch1) {
                    formatter.format("%d from ", geneInfo[j].ch1 - geneInfo[j].ch0);
                    formatter.format(Misc.nonnegIn8Chars(gtreeCoalescences[j].coalescences.get(geneInfo[j].ch0).height) + " to ");
                    formatter.format(Misc.nonnegIn8Chars(gtreeCoalescences[j].coalescences.get(geneInfo[j].ch1-1).height) + ";  ");
                }
            }
            return s.toString();
        }


        public void setLft(int lft) { this.lft = lft; }
        public void setRgt(int rgt) { this.rgt = rgt; }
        public void setAnc(int anc) { this.anc = anc; }
        public void setHeight(double height) { this.height = height; }
        public void setBNode(Node bNode) { /* nothing to do */ }

        public void setTipUnionAndID(String id) {
            this.ID = id;
            union.replaceWith(bindings.tipUnionFromSmcID(id));
        }

        //  This is used before call to tree2pionodes(), subtree2pionodes() which expect nodes to be 'empty'
        void reset() {
            anc = -1;
            lft = -1;
            rgt = -1;
            height = -1.0;
            ID = "";
            union.reset();
            for (int j = 0; j < gtreeCoalescences.length; j++) {
                geneInfo[j].reset();
            }
        }


        // for store()
        void replaceWith(PopsIONode x) {
            anc = x.anc;
            lft = x.lft;
            rgt = x.rgt;
            height = x.height;
            union.replaceWith(x.union);
            ID = x.ID;
            for (int j = 0; j < x.geneInfo.length; j++) {
                geneInfo[j].replaceWith(x.geneInfo[j]);
            }
        }

    } // end PopsIONode





    /***************************************************************************/
    /************************ Now the main class ***************************/
    /***************************************************************************/



    // constructor which reserves memory
    public PIOProbabilitySMCTree(TreeInterface tree, Bindings bindings, int [] geneTreeTipCounts) {
        this.tree = tree;
        this.bindings = bindings;
        int nTips = bindings.smcTreeTipCount();
        assert nTips == tree.getTaxonset().getTaxonCount();
        assert nTips == tree.getLeafNodeCount();
        int nNodes = 2*nTips-1;
        int nGTrees = geneTreeTipCounts.length;
        pioSNodes = new PopsIONode[2*nTips-1];
        for (int n = 0; n < nNodes; n++) {
            pioSNodes[n] = new PopsIONode(nGTrees, nTips);
        }
        gtreeCoalescences = new GTreeCoalescences[nGTrees];
        for (int j = 0; j < nGTrees; j++) {
            gtreeCoalescences[j] = new GTreeCoalescences(j, geneTreeTipCounts[j] - 1);
        }
    }


    // (re)create this tree from beast tree when it changes
    public void update() {
        for (int n = 0; n < pioSNodes.length; n++) {
            pioSNodes[n].reset();
        }
        UnionNode.Abstract.beastTreeToUnionTree(pioSNodes, tree);
    }



    // for store()
    public void replaceWith(PIOProbabilitySMCTree x) {
        assert tree == x.tree;
        assert bindings == x.bindings;
        for (int n = 0; n < pioSNodes.length; n++) {
            pioSNodes[n].replaceWith(x.pioSNodes[n]);
        }
        for (int j = 0; j < gtreeCoalescences.length; j++) {
            gtreeCoalescences[j].replaceWith(x.gtreeCoalescences[j]);
        }
    }



    /********************* debug **************************************/


    // Called from Checks, originally from PIOMSCoalescentDistribution.
    // In debug mode this checks that a node in a gene tree has been added.
    public void debugCheckCoalescenceIsThere(double height, BitUnion union, int j) {

        int nn = UnionNode.Abstract.nodeIndexOfUnionInSubtree(pioSNodes, pioSNodes.length-1, union);
        PopsIONode node = pioSNodes[nn];
        if (node.height > height) {
            System.out.println("BUG in debugCheckCoalescenceIsThere(). A gene tree is incompatible during a debug check for something else!");
        }
        assert (node.height <= height);
        while (node.anc >= 0  &&  pioSNodes[node.anc].height <= height) {
            node = pioSNodes[node.anc];
        }

        boolean found = false;
        for (int i = node.geneInfo[j].ch0; i < node.geneInfo[j].ch1; i++) {
            if (height == gtreeCoalescences[j].coalescences.get(i).height) { // exact test is deliberate
                found = true;
            }
        }
        if (!found) {
            System.out.println("BUG found by debugCheckCoalescenceIsThere(). Coalescence info not updated properly.");
        }
        assert found;
    }



    public String toString() {
        return UnionNode.Abstract.treeAsText(pioSNodes, " nlin coalheights");
    }

    /**************** Compatibility and P(smcCoalescent) calculations ***************/

    /*
      * Called from PIOProbabilityGeneTree (and originally from PIOMSCoalescentDistribution)
      * to check if a node in a gene tree is compatible with the smcTree.
      */
    public boolean coalescenceIsCompatible(double height, BitUnion union) {
        int nn = UnionNode.Abstract.nodeIndexOfUnionInSubtree(pioSNodes, pioSNodes.length-1, union);
        return (pioSNodes[nn].height <= height);
    }


    /*
    * Called from PIOProbabilityGeneTree, originally from PIOMSCoalescentDistribution,
    * to remove gene-specific coalescent information
    * from list of coalescences. Required before calls to recordCoalescence.
    */
    public void clearCoalescences(int j) {
        gtreeCoalescences[j].clearCoalescences();
    }



    /*
      * Called from PIOProbabilityGeneTree, originally from PIOMSCoalescentDistribution,
      * to add a coalescence, that is,
      * a gene tree node (identified by height+union) from gene tree j,
      * to a list of coalescences.
      */
    public void recordCoalescence(double height, BitUnion union, int j) {
        gtreeCoalescences[j].addCoalescence(height, union);
    }



    /*
      * Called from PIOProbabilityGeneTree, originally from PIOMSCoalescentDistribution
      * to accumulate information for all genes at nodes of species tree.
      */
    public void fillInCoalCountsIntensities(double maxGeneTreeHeight, double coalFactor, int j) {
        gtreeCoalescences[j].sortCoalescences();
        gtreeCoalescences[j].setPIONodeCoalHeights();
        recordSubtreeLineageCounts(pioSNodes.length - 1, j);
        fillInSubtreeCoalCountsIntensities(pioSNodes[pioSNodes.length - 1], maxGeneTreeHeight, coalFactor, j);
    }


    /*
      * Called from PIOMSCoalescentDistribution to find P(smcCoalescent)
      * by recursing over branches.
      */
    public double logProbAllGeneTreesInSMCTree(InverseGammaMixture igm, double popPriorScale) {
        return logProbAllGeneTreesInSMCSubtree(pioSNodes[pioSNodes.length - 1], igm, popPriorScale);
    }




    /*********************************************************************/
    /************************ private ************************************/
    /*********************************************************************/

    /****************** private: general/basic stuff ************************/


    // sorts the coalescences so the list is in convenient order to update
    // field in PIONodes[].
    private static final Comparator<Coalescence> COALESCENCE_ORDER = new Comparator<Coalescence>() {
        public int compare(Coalescence a, Coalescence b) {
            if (a.nn != b.nn) {
                return a.nn < b.nn ? -1 : 1;
            }
            return Double.compare(a.height, b.height);
        }
    };



    /**************** private: compatibility and probability calculations **********************************/

    private void recordSubtreeLineageCounts(int n, int j) {
        if (pioSNodes[n].lft < 0) {
            pioSNodes[n].geneInfo[j].nlineages = bindings.nLineagesFromSMCID(pioSNodes[n].ID);
        } else {
            pioSNodes[n].geneInfo[j].nlineages = 0;

            int lft = pioSNodes[n].lft;
            recordSubtreeLineageCounts(lft, j);
            int ncoalslft = pioSNodes[lft].geneInfo[j].ch1 - pioSNodes[lft].geneInfo[j].ch0;
            int nuncoalescedlft = pioSNodes[lft].geneInfo[j].nlineages - ncoalslft;
            pioSNodes[n].geneInfo[j].nlineages += nuncoalescedlft;

            int rgt = pioSNodes[n].rgt;
            recordSubtreeLineageCounts(rgt, j);
            int ncoalsrgt = pioSNodes[rgt].geneInfo[j].ch1 - pioSNodes[rgt].geneInfo[j].ch0;
            int nuncoalescedrgt =  pioSNodes[rgt].geneInfo[j].nlineages - ncoalsrgt;
            pioSNodes[n].geneInfo[j].nlineages += nuncoalescedrgt;
        }
    }




    private void fillInSubtreeCoalCountsIntensities(PopsIONode node, double maxGeneTreeHeight, double coalFactor, int j) {
        if (node.lft >= 0) {
            fillInSubtreeCoalCountsIntensities(pioSNodes[node.lft], maxGeneTreeHeight, coalFactor, j);
            fillInSubtreeCoalCountsIntensities(pioSNodes[node.rgt], maxGeneTreeHeight, coalFactor, j);
        }

        int ch0 = node.geneInfo[j].ch0;
        // find minus the log of p_j^(-k_jb), that is, k_jb log(p_j)
        int k = node.geneInfo[j].ch1 - ch0;
        node.geneInfo[j].minusLogCoalFactor = Math.log(coalFactor) * k;

        // find (1/p_j) * sum(i=0 to k_jb) c_jbi ((n_jb-i) choose 2)
        double [] t = new double[k + 2];
        t[0] = node.height;
        for (int i = 0; i < k; i++) {
            t[i + 1] = gtreeCoalescences[j].coalescences.get(ch0 + i).height;
        }
        t[k+1] = (node.anc < 0) ? maxGeneTreeHeight : pioSNodes[node.anc].height;
        int n = node.geneInfo[j].nlineages;
        node.geneInfo[j].coalintensity = 0.0;
        for (int i = 0; i <= k; i++) {
            node.geneInfo[j].coalintensity += (t[i + 1] - t[i]) * 0.5 * (n - i) * (n - i - 1) / coalFactor;
        }
    }




    private double logProbAllGeneTreesInSMCSubtree(PopsIONode node, InverseGammaMixture igm, double sigma) {
        // the names q_b, gamma_b, logMinus_r_b are to make code and maths match better
        int q_b = 0;
        double gamma_b = 0.0;
        double logMinus_r_b = 0.0;
        for (int j= 0; j < node.geneInfo.length; j++) {
            q_b += node.geneInfo[j].ch1 - node.geneInfo[j].ch0;
            logMinus_r_b += node.geneInfo[j].minusLogCoalFactor;
            gamma_b += node.geneInfo[j].coalintensity;
        }
        double [] lambdas = igm.getWeights();
        double [] alphas = igm.getAlphas();
        double [] betas = igm.getBetas();
        double [] logProbCpts = new double[lambdas.length];

        for (int c = 0; c < lambdas.length; c++) {
            logProbCpts[c] = 0.0;
            double sigmabeta_c = sigma * betas[c];
            logProbCpts[c] += Math.log(lambdas[c]);
            logProbCpts[c] += alphas[c] * Math.log(sigmabeta_c);
            logProbCpts[c] -= (alphas[c] + q_b) * Math.log(sigmabeta_c + gamma_b);
            logProbCpts[c] += lnRatioGammas(alphas[c], q_b);
        }
        double logP = logSumExp(logProbCpts);
        logP -= logMinus_r_b;

        if (node.lft >= 0) {
            logP += logProbAllGeneTreesInSMCSubtree(pioSNodes[node.lft], igm, sigma);
            logP += logProbAllGeneTreesInSMCSubtree(pioSNodes[node.rgt], igm, sigma);
        }
        return logP;
    }



    // returns log(Gamma(a+n)/Gamma(a))
    private double lnRatioGammas(double a, int n) {
        double x = 0.0;
        for (int i = 0; i < n; i++) {
            x += Math.log(a + i);
        }
        //double xx = GammaFunction.lnGamma(a + n) - GammaFunction.lnGamma(a);
        //assert Math.abs(xx-x) < 1e-8;
        return x;
    }


    // Returns log(sum_i e^(x_i)) without over/under flow
    private double logSumExp(double x[]) {
        double maxx = Double.NEGATIVE_INFINITY;
        for (double d : x) {
            if (d > maxx) { maxx = d; }
        }
        double sum = 0.0;
        for (double d : x) {
            sum += Math.exp(d-maxx);
        }
        return maxx + Math.log(sum);
    }






}
